#include "config.h"

#include <sys/types.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>

#include <openssl/sha.h>
#include <openssl/evp.h>
#include <openssl/md5.h>

#include "bloom.h"

#include <getopt.h>
#include <ctype.h>

const char *progname = "regression";
nsrl_bloom *b;  //bloom filter
int opt_x = 0;
int opt_just = 0;

const char *hex2string(char *buf,int buflen,unsigned char *md,int len)
{
    const char *bufstart = buf;
    while(len>0 && buflen+1>0){
	sprintf(buf,"%02x",md[0]);
	md ++;
	buf +=2;
	buflen -= 2;
	len --;
    }
    return bufstart;
}

void string_increment(char *buf){
	int length = strlen(buf);
	char *walker;
	char firstchar = '0';
	firstchar = firstchar-48;
	char lastchar = 'z';
	lastchar = lastchar+5;
	walker = &buf[length-1];  //start at end of string
	while(true){	
		if(*walker == lastchar){*walker = (firstchar+1); walker --;}
		else{ *walker = (*walker)++; return;}	//increment last character
		if(walker == (buf - sizeof(char))){
			walker = &buf[length];
			for(int i = 0; i<length; i++){
				buf[length] = buf[length-1];	
			}
			buf[0] = firstchar+1;
			return;	
		//	walker = &buf[length-1];return;
		} //return from left end to right end
	}
}

void usage(){
    printf("Usage: %s [options] {bloomfile|:SEED:} \n", progname);
    printf(" Specify :S:M:k to create an in-memory bloom filter and pre-seed it with the input\n");
    printf(" Specify :E:M:k to create and use an in-memory bloom filter\n");
    printf("         m = 2^M\n");
    printf("     -q         Quiet output.\n");
    printf("     -b count   Bruteforce from seed looking for a match in filter. 0 for no count.\n");
    printf("     -x         don't look anything up in the bloom filter.\n");
    printf("     -i         Displays bloomfile information.\n");
    printf("     -j         Just bring the bloom filter into memory, then quit\n");
    printf("\n input selection:\n");
    printf("     -s string  Seed string.  Default is empty string. ("")\n");
    printf("     -n int     Number of hashes to generate and query.\n");
    printf("     -f filename Filename to output generated hashed to.\n");
    printf("\n");
    printf("     -h         Shows this usage.\n");
}

/* Genprime implements a simple prime-number based random number generator */
void genprime(int count,unsigned char buf[20])
{
    /* Five big primes */
    uint32_t prime[5] = {192991619, 192991627, 192991651, 192991663, 192991679};
    uint32_t *num = (uint32_t *)buf;
    for(uint32_t i=0;i<5;i++){
	num[i] = count * prime[i];
    }
}

int main(int argc,char **argv)
{
    char message[1024];
    strncpy(message, "", 1);
    int ch=0; //for switch
    int info = 0;  //to print info
    int count = 1000000;  //how many iterations
    int	found = 0; 
    int quiet = 0;
    int brute = 0;
    int brutecount = 0;
 	unsigned char sha1[20];
    char *out_filename = 0;
    FILE *outfile = 0;
    const char *bloomfile = 0;//where bloom gets loaded from

    while((ch = getopt(argc, argv, "t:f:qs:mn:b:ihpxj")) != -1){
        switch(ch){
	case 'i': info = 1; break;
	case 'b': brute = 1; brutecount = atoi(optarg); break;
	case 'n': count = atoi(optarg); break;
	case 's': strncpy(message,optarg,40); found = 0; break;
	case 'q': quiet = 1; break;
	case 'f': out_filename = optarg; break;
	case 'x': opt_x = 1;break;
	case 'j': opt_just = 1;break;
	default:
	case 'h': usage(); exit(0);
        }//end switch
    }//end while loop for processing options

    argc -= optind;  //account for arguments handled
    argv += optind;  //walk to end of args list

    if(argc<1 && !out_filename) usage();
    if(out_filename){
	outfile = fopen(out_filename,"w");
	if(!outfile) {
	    exit(1);
	}
    }

    argc--;
    bloomfile = *argv++;

    if(bloomfile[0]==':'){
	char command;
	int M;
	int k;
	sscanf(bloomfile,":%c:%d:%d",&command,&M,&k);
	if(command!='S' && command!='E') errx(1,"Invalid bloom filter creation spec.");
	b = nsrl_bloom_alloc();
	if(nsrl_bloom_create(b,0,160,M,k,"memory")){
	    errx(1,"Cannot create bloom filter with M=%d and k=%d",M,k);
	}
	if(command=='S'){
	    /* Seed the bloom filter */
	    for(int i=0; i<count; i++){
		unsigned char hashbuf[20];
		genprime(i,hashbuf);
		nsrl_bloom_add(b, hashbuf);
	    }
	}
    }
    else{
	if(access(bloomfile,R_OK)){
	    perror(bloomfile);
	    fprintf(stderr,"Invalid Bloom file.\n");
	    exit(1);
	}
	b = nsrl_bloom_alloc();
	if(nsrl_bloom_open(b,bloomfile,O_RDONLY)){
	    err(1,"nsrl_bloom_open(%s)",bloomfile);
	}
    }

    if(opt_just){
      unsigned int counts[256];
      nsrl_calc_histogram(b,counts);
      printf("%s loaded into RAM\n",bloomfile);
      exit(0);
    }
    
    if(info){
        char ibuf[4096];
        nsrl_bloom_info(ibuf, sizeof(ibuf), b);
        fputs(ibuf, stdout);
    }

    /* The real regression loop */
    if(brute){
    	//need some logic here to ensure the string is large enough
    	//we get 89 iterations per character
    /*	int counter = strlen(message);
    	int charsgiven = 1;
    	while (counter){ charsgiven=charsgiven*89; counter--;}
    	int brutemath = brutecount;
    	int charsneeded = 0;
    	int addlchars = 0;
    	while (brutemath >0){brutemath=brutemath/89; charsneeded++;}
    	if(strlen(message) < charsneeded){
    		printf("%d chars needed, %d chars given\n", charsneeded, strlen(message));
    		addlchars = charsneeded - strlen(message);
    		char newmessage[strlen(message)];
    		strcpy(newmessage, message);
    		message[0] = '\0';
    		while(charsneeded>0){
    			strcat(message, " ");
    			charsneeded--;	
    		}
    		strcat(message, newmessage);
    	}
    	I don't like the way I did that
    	*/
    	char nmessage[40];
    	int i = 0;
    	int bruteloop = 0;
    	if(brutecount ==0){bruteloop =1;}
    	while((i<brutecount) or bruteloop){
    		SHA1((u_char *)message,strlen(message),sha1);
    		int qval = nsrl_bloom_query(b, sha1);	
    		hex2string(nmessage, 40, sha1, 20);
	    	nmessage[39] = '\0';	
    		if(!quiet){
	    		if(quiet < 1){printf("%s sha1: %s is %s\n",message,nmessage,qval ? "present" : "absent");}
			}
			if(qval){
				printf("%s sha1: %s is %s\n",message,nmessage,qval ? "present" : "absent");
				found++;
				if(bruteloop){break;}
			}
			string_increment(message);
			i++;
    	}
    }
    
    else{
    	for(int i=0; i<count; i++){
		unsigned char hashbuf[20];
    	genprime(i,hashbuf);
		if(opt_x) continue;		/* don't lookup in the bloom filter */
		int qval = nsrl_bloom_query(b, hashbuf);
		if(!quiet){
	    	hex2string(message, 40, hashbuf, 20);
	    	message[40] = '\0';
	    	printf("%s is %s\n",message,qval ? "present" : "absent");
		}
		if(qval) found++;
    	}
    }
    printf("%d out of %d found.  %f rate.\n", found, brute ? brutecount : count, (100*(float)found/(float)count));
    nsrl_bloom_print_info(b);
    nsrl_bloom_print_usage();
    return 0;
}
